#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>

#include "tifmaker.h"


FILE *fpo,*fpi;
int skip_byte=0;	/* skip byte count  */

#define MP 3
			 /*------------------- TIFF information -----------------------*/
unsigned short int		Mp=1;			/* number of Image plane */
unsigned short int      ByteOrder=MOTOR;	   /* Byte order 0:Intel 1:Motorolla */

unsigned long int      ImageWidth=0;           /* 100:Pixel Width   */
unsigned long int      ImageLength=0;          /* 101:Pixel Height   */
unsigned short int      BitsPerSample;	       /* 102:Bits Per Pixel  */
unsigned short int      Compression=1;          /* 103:Compression Type       */
unsigned short int      PhotoInter=1;           /* 106:Image Type  */
unsigned short int      FillOrder=1;            /* 10a:Bit Filling Order     */
unsigned long			StripOffsets[MAXstrip]; /* 111:Image data Pointer */
unsigned short int      SamplesPerPixel=1;      /* 115:Data Per plane (1 or 3) */
unsigned short int      Orientation=1;          /* 112:Orientation(1:TopLeftj */
unsigned short int      BgrOrder=0;

unsigned long 			XResolution=180;       /* x-direction dpi      */
unsigned long			YResolution=180;       /* y-direction dpi      */

char mode_skip_byte=' ';
char mode_ImageWidth=' ';
char mode_ImageLength=' ';
char mode_BitsPerSample=' ';
char mode_PhotoInter=' ';
char mode_FillOrder=' ';
char mode_Orientation=' ';


int Write_Short(int x)
{
	uchar tmp;
	Uni.i[0]=x;
	if(ByteOrder==MOTOR){
		tmp=Uni.c[0]; Uni.c[0]=Uni.c[1]; Uni.c[1]=tmp;
	}
	return fwrite(Uni.i,2,1,fpo);
}
int	Write_Long(long x)
{
	uchar tmp;
	Uni.l=x;
	if(ByteOrder==MOTOR)  {          /*  Swap byte        */
		tmp=Uni.c[0];Uni.c[0]=Uni.c[3];Uni.c[3]=tmp;
		tmp=Uni.c[1];Uni.c[1]=Uni.c[2];Uni.c[2]=tmp;
	}
	return fwrite(&Uni.l,4,1,fpo);
}

long pack_bits(char buf[])
{
	int xn,xlen,i,j,k;
	xn=8/BitsPerSample;
	xlen=(ImageWidth+7)/xn;
	if(PhotoInter==2)	xlen=xlen*3;
	i=0;
	while(1){
		j=i+1;
		while(buf[i]==buf[j] && j<xlen)		j++;
		k=j-i;
		if(k>1){
			j=-k+1;
			putc(j,fpo);	putc(buf[i],fpo);
			i+=k;
			if(i>=xlen)	break;
		}else{
			j=i+2;
			while(buf[j-1]!=buf[j] && buf[j]!=buf[j+1] && j<xlen)	j+=1;
			k=j-i; j=k-1;
			putc(j,fpo);
			for(j=0;j<k;j++){
				putc(buf[i++],fpo);
				if(i>=xlen)	break;
			}
			if(i>xlen)	break;
		}
	}
	return ftell(fpo);
}

void Write_TagS(int id,int dat)
{
	Write_Short(id);
	Write_Short(TIFF_SHORT);
	Write_Long(1L);
	Write_Short(dat);Write_Short(0);
}

void Write_TagSs(int id,long adr,int n)
{
	Write_Short(id);
	Write_Short(TIFF_SHORT);
	Write_Long((long)n);
	Write_Long(adr);
}

void Write_TagL(int id,long dat)
{
	Write_Short(id);
	Write_Short(TIFF_LONG);
	Write_Long(1L);
	Write_Long(dat);
}
void Write_TagLs(int id,long adr,int n)
{
	Write_Short(id);
	Write_Short(TIFF_LONG);
	Write_Long((long)n);
	Write_Long(adr);
}
void Write_TagR(int id,long adr)
{
	Write_Short(id);
	Write_Short(TIFF_RATIONAL);
	Write_Long(1L);
	Write_Long(adr);
}
void Write_TagStrings(int id,long adr)
{
	Write_Short(id);
	Write_Short(TIFF_RATIONAL);
	Write_Long(1L);
	Write_Long(adr);
}


/* Save Tiff-Image File as gray-scale */
void mono_save(void)
{
	char buf[20];
	int i,TagCnt;
	long ImageSize,ImageOffset,TagSize,TagOffset,ExtSize,ExtOffset;
	int w_byte;	/* byte count per 1line */
	char *work;

	StripOffsets[0]=8L;     /* 111:Image data pointer */
	TagCnt=12;
	TagOffset=8L; 					 TagSize=TagCnt*12+2+4;
	ExtOffset=TagOffset+TagSize;     ExtSize=8*2;
	ImageOffset=ExtOffset+ExtSize;
	ImageSize=(long)ImageWidth*ImageLength*BitsPerSample/8L;

	if(ByteOrder==INTEL)	buf[0]=buf[1]='I';
	else					buf[0]=buf[1]='M';
	fwrite(buf,1,2,fpo);
	Write_Short(42);
	Write_Long(TagOffset);

	Write_Short(TagCnt);
	Write_TagS(T_SubfileType,1);
	Write_TagL(T_ImageWidth,ImageWidth);
	Write_TagL(T_ImageLength,ImageLength);
	Write_TagS(T_BitsPerSample,BitsPerSample);
	Write_TagS(T_Compression,Compression);
	Write_TagS(T_PhotoInter,PhotoInter);
	Write_TagS(T_FillOrder,FillOrder);
	Write_TagS(T_Orientation,Orientation);

	Write_TagL(T_StripOffsets,ImageOffset);
	Write_TagS(T_SamplesPerPixel,SamplesPerPixel);
	Write_TagR(T_XResolution,ExtOffset);
	Write_TagR(T_YResolution,ExtOffset+8L);
	Write_Long(0L);
	Write_Long(XResolution);Write_Long(1L);
	Write_Long(YResolution);Write_Long(1L);

	w_byte=ImageSize/ImageWidth;	
	work=malloc(w_byte);
	if(work==NULL){
		fprintf(stderr,"Not enough memory\n");
		exit(EXIT_FAILURE);
	}
	if(Compression==0 || Compression==1){
		while((i=fread(work,sizeof(char),w_byte,fpi))>0){
			if(fwrite(work,sizeof(char),i,fpo)<=0){
				fprintf(stderr,"Can't write to file\n");
				free(work);
				exit(EXIT_FAILURE);
			}
		}
	}else if(Compression==32773){	/* PackBits compression */
		while((i=fread(work,sizeof(char),w_byte,fpi))>0){
			pack_bits(work);
		}
	}
	free(work);
}


/* Save Tiff-Image File as full-color Image (Single plane) */
void col_save1(void)
{
	char buf[20];
	char bdata;
	int i,j,TagCnt;
	long ImageSize,ImageOffset,TagSize,TagOffset,ExtSize,ExtOffset;
	int w_byte;	/* byte count per line */
	char *work;

	StripOffsets[0]=8L;     /* 111:Image data pointer */
	SamplesPerPixel=3;      /* 115:data per PIXEL (3:RGBcolor) */

	TagCnt=13;
	TagOffset=8L; 					 TagSize=TagCnt*12+2+4;
	ExtOffset=TagOffset+TagSize;     ExtSize=8*2;//ExtSize=8*2+4*3;
	ImageOffset=ExtOffset+ExtSize;
	ImageSize=(long)ImageWidth*ImageLength*BitsPerSample/8L*3;

	if(ByteOrder==INTEL)	buf[0]=buf[1]='I';
	else					buf[0]=buf[1]='M';
	fwrite(buf,1,2,fpo);
	Write_Short(42);
	Write_Long(TagOffset);

	Write_Short(TagCnt);
	Write_TagS(T_SubfileType,1);
	Write_TagL(T_ImageWidth,ImageWidth);
	Write_TagL(T_ImageLength,ImageLength);
	Write_TagS(T_BitsPerSample,BitsPerSample);
	// Write_TagSs(T_BitsPerSample,ExtOffset+16L,3);
	Write_TagS(T_Compression,Compression);
	Write_TagS(T_PhotoInter,PhotoInter);
	Write_TagS(T_FillOrder,FillOrder);
	Write_TagS(T_Orientation,Orientation);

	Write_TagL(T_StripOffsets,ImageOffset);
	Write_TagL(T_StripByteCounts,ImageSize);

	Write_TagS(T_SamplesPerPixel,SamplesPerPixel);
	Write_TagR(T_XResolution,ExtOffset+6);
	Write_TagR(T_YResolution,ExtOffset+8L);
	Write_Long(0L);
	Write_Long(XResolution);Write_Long(1L);
	Write_Long(YResolution);Write_Long(1L);

	// Write_Short(BitsPerSample);
	//Write_Short(BitsPerSample);
	//Write_Short(BitsPerSample);


	w_byte=ImageSize/ImageWidth;	/* byte count per line */
	work=malloc(w_byte);
	if(work==NULL){
		fprintf(stderr,"Not enough memory\n");
		exit(EXIT_FAILURE);
	}
	if(Compression==0 || Compression==1){
		while((i=fread(work,sizeof(char),w_byte,fpi))>0){
			if(BgrOrder){
				for(j=0;j<w_byte-2;j+=3){
					bdata=work[j];
					work[j]=work[j+2];
					work[j+2]=bdata;
				}
			}
			if(fwrite(work,sizeof(char),i,fpo)<=0){
				fprintf(stderr,"Can't write to file\n");
				free(work);
				exit(EXIT_FAILURE);
			}
		}
	}else if(Compression==32773){	/* PackBits compression */
		while((i=fread(work,sizeof(char),w_byte,fpi))>0){
			pack_bits(work);
		}
	}
	free(work);
}




void usage(void)
{
	printf("\nTIFMAKER Ver 0.30 ,\t\tCopyright (c) 1994-99 by Horie-ms\n");
	printf("@Convert from Binary-file to Tiff-Image-file\n");
	printf("  TIFMAKER@Sorce-File@[path]mDestination-file[.tif]] [Option]\n\n\n");

	printf("    Source-File Options\n");
	printf("\t/W###\tX-pixels(horozontal)  (Default: /W512)\n");
	printf("\t/L###\tY-pixels(vertical)    (Default: Auto\n");
	printf("\t/D#\tBits Per Pixel          (Default: /D8)\n");
	printf("\t/P#\tPixel Data Format  (Default: /P1)\n");
	printf("\t           0:Gray Image (0 is White)\n");
	printf("\t           1:Gray Image (0 is Bklack)\n");
	printf("\t           2:RGB Color Image (Packed)\n");
	printf("\t           3:RGB Color Image (Plane)\n");
	printf("\t/F#\tBit Filling Order  (Default /F1)\n");
	printf("\t           1 : Left to Right\n");
	printf("\t           2 : Right to Left\n");
	printf("\t/O#\tOrientation  (Default /O1)\n");
	printf("\t           1 : Top Leftt\n");
	printf("\t           2 : Top Right\n");
	printf("\t           3 : Bottom Right\n");
	printf("\t           4 : Bottom Left\n");
	printf("\t/S##\tSkip BytesiHeader Byte Count) (Default: Auto)\n");
	printf("\t/R##\tColor Order (Default: 0)\n");
	printf("\t           0 : RGB\n");
	printf("\t           1 : BGR\n\n");

	printf("@@Destination-Image Options\n");
	printf("\t/B#\tByte Order (Default /BMj\n");
	printf("\t           I(0):Intel\n");
	printf("\t           M(1):Motorolla\n");
	printf("\t/C#\tCompression Type iDefault /CNj\n");
	printf("\t           N:None\n");
	printf("\t           P:PackBits Compression\n");
	printf("\t/X###\tHorozontal dpi@@iDefault /X180j\n");
	printf("\t/Y###\tVertical dpi@@  iDefault /Y180j\n\n");

	printf("Memo\n@1)Option settings can be changed according to the input file extension by the definition file[TIFFMAKER.dfn].\n");
	printf("\n  2)The following options can set a value from the information of the position of the input file.\n");
	printf("\n      /S /W /L /D /P /F /O\n");
	printf("\n      Ex. /W16w  Read Word data from the 16th file position\n");
	printf("\n      As for the suffix, the following letter is permitted.\n");
	printf("\n         b:Byte,w:Word,d:DoubleWord\n");

	exit(EXIT_FAILURE);
}

void get_fname(char argc,char *argv[],char *i_name,char *o_name)
{
	int i;
	char sw;
	char drive[_MAX_DRIVE],dir[_MAX_DIR],file[_MAX_FNAME],ext[_MAX_EXT];
	char dumy[_MAX_FNAME];

	for(i=1;i<argc;i++){
		sw=argv[i][0];
		if(sw!='-' && sw!='/'){
			if(strlen(i_name)==0)	strcpy(i_name,argv[i]);
			else					strcpy(o_name,argv[i]);
		}
	}
	if(strlen(i_name)==0)	usage();
	if(strlen(o_name)==0){
		_splitpath( i_name,drive,dir,file,ext);
		strcpy(ext,".tif");
		_makepath(o_name,drive,dir,file,ext);
	}else{
		_splitpath(i_name,drive,dir,dumy,ext);
		_splitpath(o_name,drive,dir,file,ext);
		if(strlen(file)==0)	strcpy(file,dumy);
		if(strlen(ext)==0)	strcpy(ext,".tif");
		_makepath(o_name,drive,dir,file,ext);
	}
}
void get_option(char *opt)
{
	int n;
	int num;
	char c;
	
	strupr(opt);
	if(opt[0]=='-' || opt[0]=='/'){
		switch(opt[1]){
			case 'B': if(opt[2]=='I'||opt[2]=='0')	ByteOrder=INTEL;
					  else 							ByteOrder=MOTOR;
					  break;
			case 'W': n=sscanf(&opt[2],"%d%c",&num,&c);
					  if(n>=1)	ImageWidth=num;
					  if(n>=2)	mode_ImageWidth=tolower(c);
					  break;
			case 'L': n=sscanf(&opt[2],"%d%c",&num,&c);
					  if(n>=1)	ImageLength=num;
					  if(n>=2)	mode_ImageLength=tolower(c);
					  break;
			case 'D': n=sscanf(&opt[2],"%d%c",&num,&c);
					  if(n>=1)	BitsPerSample=num;
					  if(n>=2)	mode_BitsPerSample=tolower(c);
					  break;
			case 'P': n=sscanf(&opt[2],"%d%c",&num,&c);
					  if(n>=1)	PhotoInter=num;
					  if(n>=2)	mode_PhotoInter=tolower(c);
					  break;
			case 'F': n=sscanf(&opt[2],"%d%c",&num,&c);
					  if(n>=1)	FillOrder=num;
					  if(n>=2)	mode_FillOrder=tolower(c);
					  break;
			case 'C': if(opt[2]=='P')	Compression=32773;
					  else				Compression=1;
					  break;
			case 'S': n=sscanf(&opt[2],"%d%c",&num,&c);
					  if(n>=1)	skip_byte=num;
					  if(n>=2)	mode_skip_byte=tolower(c);
					  break;
			case 'X': XResolution=atol(&opt[2]);		break;
			case 'Y': YResolution=atol(&opt[2]);		break;
			case 'O': n=sscanf(&opt[2],"%d%c",&num,&c);
					  if(n>=1)	Orientation=num;
					  if(n>=2)	mode_Orientation=tolower(c);
					  break;
			case 'R': BgrOrder=atol(&opt[2]);		break;
			default:
				fprintf(stderr,"Option[%s] not exist\n",opt);
		}
	}
}

void dfn_chk(char *prog_name,char *in_file)	/* check definition file */
{
	char drive[_MAX_DRIVE],dir[_MAX_DIR],file[_MAX_FNAME],ext[_MAX_EXT];
	char dfn_name[_MAX_PATH];
	char buf[81],*p;
	FILE *fp;

	_splitpath(prog_name,drive,dir,file,ext);
	strcpy(ext,".dfn");
	_makepath(dfn_name,drive,dir,file,ext);
	fp=fopen(dfn_name,"rt");
	if(fp==NULL)	return;
	_splitpath(in_file,drive,dir,file,ext);
	while(fgets(buf,81,fp)!=NULL){
		if(buf[0]==';' || buf[0]=='*')	continue;
		if(buf[0]!='.' || _strnicmp(buf,ext,strlen(ext))==0){
			p = strtok(buf," \t");
			while(p!=NULL){
				get_option(p);
				p=strtok(NULL," \t");
			}
		}
	}
	fclose(fp);
}



void read_header(void)
{
	unsigned char bdata;
	unsigned short int wdata;
	unsigned long int ddata;

	if(mode_skip_byte=='b' || mode_skip_byte=='w' || mode_skip_byte=='d'){
		fseek(fpi,skip_byte,SEEK_SET);
		if(mode_skip_byte=='b'){
			fread(&bdata,sizeof(bdata),1,fpi);
			skip_byte=bdata;
		}else if(mode_skip_byte=='w'){
			fread(&wdata,sizeof(wdata),1,fpi);
			skip_byte=wdata;
		}else{
			fread(&ddata,sizeof(ddata),1,fpi);
			skip_byte=ddata;
		}
	}
	if(mode_ImageWidth=='b' || mode_ImageWidth=='w' || mode_ImageWidth=='d'){
		fseek(fpi,ImageWidth,SEEK_SET);
		if(mode_ImageWidth=='b'){
			fread(&bdata,sizeof(bdata),1,fpi);
			ImageWidth=bdata;
		}else if(mode_ImageWidth=='w'){
			fread(&wdata,sizeof(wdata),1,fpi);
			ImageWidth=wdata;
		}else{
			fread(&ddata,sizeof(ddata),1,fpi);
			ImageWidth=ddata;
		}
	}
	if(mode_ImageLength=='b' || mode_ImageLength=='w' || mode_ImageLength=='d'){
		fseek(fpi,ImageLength,SEEK_SET);
		if(mode_ImageLength=='b'){
			fread(&bdata,sizeof(bdata),1,fpi);
			ImageLength=bdata;
		}else if(mode_ImageLength=='w'){
			fread(&wdata,sizeof(wdata),1,fpi);
			ImageLength=wdata;
		}else{
			fread(&ddata,sizeof(ddata),1,fpi);
			ImageLength=ddata;
		}
	}
	if(mode_BitsPerSample=='b' || mode_BitsPerSample=='w' || mode_BitsPerSample=='d'){
		fseek(fpi,BitsPerSample,SEEK_SET);
		if(mode_BitsPerSample=='b'){
			fread(&bdata,sizeof(bdata),1,fpi);
			BitsPerSample=bdata;
		}else if(mode_BitsPerSample=='w'){
			fread(&wdata,sizeof(wdata),1,fpi);
			BitsPerSample=wdata;
		}else{
			fread(&ddata,sizeof(ddata),1,fpi);
			BitsPerSample=ddata;
		}
	}
	if(mode_PhotoInter=='b' || mode_PhotoInter=='w' || mode_PhotoInter=='d'){
		fseek(fpi,PhotoInter,SEEK_SET);
		if(mode_PhotoInter=='b'){
			fread(&bdata,sizeof(bdata),1,fpi);
			PhotoInter=bdata;
		}else if(mode_PhotoInter=='w'){
			fread(&wdata,sizeof(wdata),1,fpi);
			PhotoInter=wdata;
		}else{
			fread(&ddata,sizeof(ddata),1,fpi);
			PhotoInter=ddata;
		}
	}
	if(mode_FillOrder=='b' || mode_FillOrder=='w' || mode_FillOrder=='d'){
		fseek(fpi,FillOrder,SEEK_SET);
		if(mode_FillOrder=='b'){
			fread(&bdata,sizeof(bdata),1,fpi);
			FillOrder=bdata;
		}else if(mode_FillOrder=='w'){
			fread(&wdata,sizeof(wdata),1,fpi);
			FillOrder=wdata;
		}else{
			fread(&ddata,sizeof(ddata),1,fpi);
			FillOrder=ddata;
		}
	}
	if(mode_Orientation=='b' || mode_Orientation=='w' || mode_Orientation=='d'){
		fseek(fpi,Orientation,SEEK_SET);
		if(mode_Orientation=='b'){
			fread(&bdata,sizeof(bdata),1,fpi);
			Orientation=bdata;
		}else if(mode_Orientation=='w'){
			fread(&wdata,sizeof(wdata),1,fpi);
			Orientation=wdata;
		}else{
			fread(&ddata,sizeof(ddata),1,fpi);
			Orientation=ddata;
		}
	}
	if(BitsPerSample>=24){
		if(BitsPerSample>=48)	BitsPerSample=16;
		else					BitsPerSample=8;
		PhotoInter=2;		/* RGB color image */			
	}
}



int main(int argc,char *argv[])
{
	char in_name[_MAX_PATH],out_name[_MAX_PATH],tmp_name[_MAX_PATH];
	char drive[_MAX_DRIVE+10],dir[_MAX_DIR+10],file[_MAX_FNAME],ext[_MAX_EXT];
	char buf[80],dumy[_MAX_PATH];
	int i,cycle;
	ldiv_t	lx;
	struct _finddata_t fileinfo;
	long hFile;

	Mp=1;
	ByteOrder=MOTOR;	/* B:Byte Order  0:Intel 1:motorolla */
	ImageWidth=512;	    /* W:100:Image width   */
	ImageLength=0;	    /* L:101:Image height  */
	BitsPerSample=8;	/* D:102:Bits per 1PIXEL(1)  */
	PhotoInter=0;       /* P:106:Image data format  */
	FillOrder=1;        /* F:10a:bit filling order     */
	SamplesPerPixel=Mp; /* 115:samples Per Pixel(1 or 3) */
	Compression=1;      /* C:103:Compression type(1)  (32773:Packbits)     */
	skip_byte=-1;		/* S:sjip byte count */

	strcpy(in_name,"");	strcpy(out_name,"");

	get_fname(argc,argv,in_name,out_name);	/* input file name */
	dfn_chk(argv[0],in_name);				/* read dfn file */
	for(i=1;i<argc;i++)	get_option(argv[i]);	/* check argv ߼ */

	strcpy(tmp_name,in_name);
	hFile = _findfirst( tmp_name, &fileinfo );
	if(hFile==-1L){
		fprintf(stderr,"Not Exist [%s] \n",tmp_name);
		return EXIT_FAILURE;
	}
	cycle=0;
	do{
		if(strchr(tmp_name,'*')!=NULL || strchr(tmp_name,'?')!=NULL){
			_splitpath(in_name,drive,dir,file,ext);
			strcpy(file,fileinfo.name);
			_splitpath(file,dumy,dumy,buf,dumy);
			_makepath(in_name,drive,dir,buf,ext);
			_splitpath(in_name,drive,dir,file,ext);
			strcpy(ext,".tif");
			_makepath(out_name,drive,dir,file,ext);
		}
		fpi=fopen(in_name,"rb");
		if(fpi==NULL){
			fprintf(stderr,"Can't open [%s] (read)\n",in_name);
			exit(EXIT_FAILURE);
		}
		read_header();
		printf("\r%s --> %s converting..",in_name,out_name);
		fseek(fpi,skip_byte,SEEK_SET);

		fpo=fopen(out_name,"wb");
		if(fpo==NULL){
			fprintf(stderr,"Can't open [%s] (write)\n",out_name);
			exit(EXIT_FAILURE);
		}
		if(ImageLength==0||skip_byte<0){
			lx=ldiv(filelength(fileno(fpi)),ImageWidth/(8/BitsPerSample));
			if(ImageLength==0)	ImageLength=lx.quot;
			if(skip_byte<0)		skip_byte=lx.rem;
		}
		if(PhotoInter==2)	  	col_save1();
		else					mono_save();
		fclose(fpo);
		fclose(fpi);

		printf("\r%s --> %s convered!            \n",in_name,out_name);
		cycle+=1;
	}while( _findnext( hFile, &fileinfo ) == 0 );

	 _findclose( hFile );
	return EXIT_SUCCESS;
}

